"""Convert a pandoc string to ansi
"""
import urllib
import os
import catimage
import art

import pygments
from pygments.formatters import Terminal256Formatter
from pygments.lexers import get_lexer_by_name

from catpandoc import processpandoc
from catpandoc.pandoc2plain import Pandoc2Plain


class Pandoc2Ansi(Pandoc2Plain):
	"""Convert a pandoc string to ansi. Inherits from Pandoc2Plain (no need
	for identical methods - down with code duplication...)
	"""
	def __init__(self, width=80, padding=0, baseColour=(4, 0, 4)):
		self.content = []
		self.width = width - padding * 2
		self.padding = padding
		self.baseColour = baseColour
		# Constants
		self.FG = "\033[38;5;"
		self.BG = "\033[48;5;"
		self.CLR_FG = "\033[39m"
		self.CLR_BG = "\033[49m"
		self.CLR = "\033[0m"


	#########################################
	# UTIL
	#########################################

	def catImage(self, content):
		'''catImage '''
		self.print(self.CLR)
		try:
			if content.startswith("http"):
				urllib.request.urlretrieve(content, "dowloadedImage")
				content = "dowloadedImage"
			try:
				self.print(catimage.generateHDColour(os.getcwd() + os.sep + content,
				self.width))
			except FileNotFoundError:
				self.print(content)
		except urllib.error.HTTPError:
			self.print(content)
		self.print(self.CLR, end="")


	#########################################
	# BLOCK
	#########################################

	def header(self, content):
		'''Process a header '''
		self.print("\n", end="")
		concatContent = ""
		headColour = [0,
		self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (1, 1, 1))),
		self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (1, 0, 0))),
		self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (0, 0, 0))),
		self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (0, 0, -1))),
		self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (-1, -1, -1))),
		self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (-1, -1, -2)))]
		if content[0] > 3:
			self.print(self.FG + headColour[content[0]], end="")
			for part in content[-1]:
				processpandoc.processInline(part, self)
			self.print(self.CLR, end="")
		else:
			concatContent = "".join([processpandoc.toPlaintext(part) for part in content[-1]])
			if any(char in concatContent for char in ["q", "y", "p", "g", "j"]):
				limit = 1
			else:
				limit = 2
			if content[0] == 1:
				if len(concatContent) > self.width / 7:
					concatContent = concatContent[:int(self.width / 7)]
				self.print(self.FG + headColour[content[0]] +
				"\n".join(art.text2art(concatContent, "swan").split("\n")[2:-limit-1]))
				self.print("▀"*self.width + self.CLR)
			if content[0] == 2:
				if len(concatContent) > self.width / 5:
					concatContent = concatContent[:int(self.width / 5)]
				self.print(self.FG + headColour[content[0]] +
				"\n".join(art.text2art(concatContent, "thin").split("\n")[1:-limit]))
				self.print("━"*self.width + self.CLR)
			if content[0] == 3:
				if len(concatContent) > self.width / 3:
					concatContent = concatContent[:int(self.width / 3)]
				self.print(self.FG + headColour[content[0]] +
				"\n".join(art.text2art(concatContent, "cygnet").split("\n")[1:-limit]))
				self.print("─"*self.width + self.CLR)

	def codeBlock(self, content):
		'''Process a code block  '''
		codeColour = self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (-1, -1, -1)))
		self.print(self.FG + codeColour + "\n │ " + self.CLR_FG, end="")
		try:
			self.print(self.FG + codeColour + " │ ".join([self.CLR_FG +
			pygments.highlight(line, get_lexer_by_name(content[0][1][0]), Terminal256Formatter()) +
			self.CLR for line in content[1].split("\n")]))
		except:
			self.print(self.FG + codeColour + "\n │ ".join(content[1].split("\n")) + self.CLR_FG)

	def definitionList(self, content):
		'''Process a definition list '''
		for definition in content:
			#for definition in definitionBlock:
			self.newline()
			self.print(self.FG + self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (-1, -1, -2))), end="")
			for part in definition[0]:
				processpandoc.processInline(part, self)
			self.print("\t:"+ self.CLR_FG + ":\t", end="")
			for part in definition[1]:
				for partB in part:
					processpandoc.processBlock(partB, self)


	def orderedList(self, content):
		'''Process an ordered list '''
		self.print(self.FG + self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (-1, -1, -2))), end="")
		for index, bullet in enumerate(content[1]):
			self.print("\n", index+1, end=". ")
			for point in bullet:
				if point["t"] in ["BulletList"]:
					self.print(" > ", end="")
				processpandoc.processBlock(point, self)
		self.print(self.CLR_FG, end="")

	def bulletList(self, content):
		'''Process a bulleted list '''
		self.print(self.FG + self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (-1, -1, -2))), end="")
		for bullet in content:
			for point in bullet:
				if point["t"] not in ["BulletList"]:
					self.print("\n- ", end="")
				else:
					self.print(" > ", end="")
				processpandoc.processBlock(point, self)
		self.print(self.CLR_FG, end="")

	# Table

	def blockQuote(self, content):
		'''Process a block quote '''
		print("\n" + self.FG + self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (-2, -2, -2))) +
		" ┃ ", end="")
		print("\n ┃ ".join([processpandoc.toPlaintext(block).split("\n") for block in content][0]) +
		self.CLR_FG)


	#########################################
	# INLINE
	#########################################

	# Space

	def emph(self, content):
		'''Process emphasized text '''
		for newInline in content:
			self.print("\033[3m", end="")
			processpandoc.processInline(newInline, self)
			self.print("\033[23m", end="")

	def strong(self, content):
		'''Process strong (bold) text '''
		for newInline in content:
			self.print("\033[1m", end="")
			processpandoc.processInline(newInline, self)
			self.print("\033[21m", end="")

	def strikeout(self, content):
		'''Process strikeout (crossed out) text '''
		for newInline in content:
			self.print("\033[9m", end="")
			processpandoc.processInline(newInline, self)
			self.print("\033[29m", end="")

	# Superscript

	# Subscript

	# SmallCaps

	# Quoted

	def code(self, content):
		'''Process code '''
		self.print(self.FG + self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (-1, -1, -1))) +
		" │" + content[1], end="│ " + self.CLR_FG)

	# Math

	# Note

	# Span

	# Image

	# Link

	def hr(self):
		'''Process a hr '''
		self.print(self.FG + self.ansiFromAnsiRgb(self.varBaseC(self.baseColour, (0, 0, 0))) +
		"─"*self.width + self.CLR_FG)


	def ansiFromAnsiRgb(self, colour):
		'''Generates ansi based on reduced rgb ie in range 0-5 '''
		r, g, b = (colour)
		return str(int(16 + min(max(r, 0), 5) * 36 + min(max(g, 0), 5) * 6 + min(max(b, 0), 5))) + "m"

	def varBaseC(self, base, modify):
		'''Generates a colour based on the theme colour '''
		newC = []
		for index in range(3):
			newC.append(base[index] + modify[index])
		return newC



"""Holds functions to iterate through the pandoc json

Block
	Plain
	Para
	LineBlock
	CodeBlock
	RawBlock
	BlockQuote
	OrderedList
	BulletList
	DefinitionList
	Header
	HorizontalRule
	Table
	Div
	Null


Inline
	Str
	Emph
	Strong
	Strikeout
	Superscript
	Subscript
	SmallCaps
	Quoted
	Cite
	Code
	Space
	SoftBreak
	LineBreak
	Math
	RawInline
	Link
	Image
	Note
	Span


"""

import re
import emoji

def toPlaintext(inline):
	"""Convert an inline or block to plain text

	Args:
		inline (dict): An inline holds various types such as 'Space', 'Bold'
		and other Inlines

	Returns:
		string: plain text
	"""
	if inline["t"] == "Space":
		return " "
	if inline["t"] == "Str":
		# text
		return emoji.emojize(inline["c"])
	if inline["t"] in ["Emph", "Strong", "Strikeout", "Superscript", "Subscript", "SmallCaps"]:
		# inline[]
		text = []
		for newInline in inline["c"]:
			text.append(toPlaintext(newInline))
		return "".join(text)
	if inline["t"] == "Quoted":
		# quotetype, inline[]
		text = []
		for newInline in inline["c"]:
			text.append("\'" + toPlaintext(newInline) + "\'")
		return "".join(text)
	if inline["t"] == "Code":
		return " │" + inline["c"][1] + " │ "
	if inline["t"] in ["Plain", "Para"]:
		# inline[]
		text = []
		for newInline in inline["c"]:
			text.append(toPlaintext(newInline))
		return "".join(text)
	if inline["t"] in ["SoftBreak", "LineBreak"]:
		return "\n"
	if inline["t"] == "Link":
		return inline["c"][2][0]
	if inline["t"] == "RawInline":
		return inline["c"][1]
	print(inline)



def processRaw(content, pandoc2):
	"""Process raw data - this might be html

	Args:
		content ([string, string]): content type and content data
		pandoc2 (object): pandoc2XYZ object formatter
	"""
	if content[0] in ["html"]:
		if content[1].startswith("<img"):
			pandoc2.catImage(re.findall(r"src=\"(.*?)\"", content[1])[0])
		elif content[1].startswith("<!--"):
			pass
		else:
			pandoc2.print(content)
	else:
		pandoc2.print(content)

def processInline(inline, pandoc2):
	"""Do stuff for an inline object

	Args:
		inline (dict): An inline holds various types such as 'Space', 'Bold'
		and other Inlines
		pandoc2 (object): pandoc2XYZ object formatter
	"""
	if inline["t"] == "Str":
		# text
		pandoc2.print(emoji.emojize(inline["c"]), end="")
	if inline["t"] == "Emph":
		# inline[]
		pandoc2.emph(inline["c"])
	if inline["t"] == "Strong":
		# inline[]
		pandoc2.strong(inline["c"])
	if inline["t"] == "Strikeout":
		# inline[]
		pandoc2.strikeout(inline["c"])
	if inline["t"] == "Superscript":
		# inline[]
		pandoc2.superscript(inline["c"])
	if inline["t"] == "Subscript":
		# inline[]
		pandoc2.subscript(inline["c"])
	if inline["t"] == "SmallCaps":
		# inline[]
		pandoc2.smallCaps(inline["c"])
	if inline["t"] == "Quoted":
		# quotetype, inline[]
		pandoc2.quoted(inline["c"])
	if inline["t"] == "Cite":
		# citation[], inline[]
		pandoc2.cite(inline["c"])
	if inline["t"] == "Code":
		# attributes, text
		pandoc2.code(inline["c"])
	if inline["t"] == "Space":
		pandoc2.space()
	if inline["t"] == "SoftBreak":
		pandoc2.newline()
	if inline["t"] == "LineBreak":
		pandoc2.newline()
	if inline["t"] == "Math":
		# mathtype, text
		pandoc2.math(inline["c"])
	if inline["t"] == "RawInline":
		# format, text
		processRaw(inline["c"], pandoc2)
	if inline["t"] == "Link":
		# attr, inline[], text
		pandoc2.link(inline["c"])
	if inline["t"] == "Image":
		# attr, inline[], text
		pandoc2.image(inline["c"])
	if inline["t"] == "Note":
		# block[]
		pandoc2.note(inline["c"])
	if inline["t"] == "Span":
		# attr, inline[]
		pandoc2.span(inline["c"])

def processBlock(block, pandoc2):
	"""Do stuff for an block object

	Args:
		block (dict): An block holds various types such as 'Para', 'Table'
		pandoc2 (object): pandoc2XYZ object formatter
	"""
	if block["t"] == "Plain":
		# inline[]
		for inline in block["c"]:
			processInline(inline, pandoc2)
	if block["t"] == "Para":
		# inline[]
		for inline in block["c"]:
			processInline(inline, pandoc2)
	if block["t"] == "LineBlock":
		# inline[][]
		pandoc2.lineBlock(block["c"])
	if block["t"] == "CodeBlock":
		# language, text
		pandoc2.codeBlock(block["c"])
	if block["t"] == "RawBlock":
		# type, text
		processRaw(block["c"], pandoc2)
	if block["t"] == "BlockQuote":
		# block[]
		pandoc2.blockQuote(block["c"])
	if block["t"] == "OrderedList":
		# attributes, block[][]
		pandoc2.orderedList(block["c"])
	if block["t"] == "BulletList":
		# block[][]
		pandoc2.bulletList(block["c"])
	if block["t"] == "DefinitionList":
		# [key, value] key=inline[] value=block[]
		pandoc2.definitionList(block["c"])
	if block["t"] == "Header":
		# int, attr, inline[]
		pandoc2.header(block["c"])
	if block["t"] == "HorizontalRule":
		pandoc2.hr()
	if block["t"] == "Table":
		# inline[], alignment[], double[], tablecell[], tablecell[][]
		# ignore. align, ignore, tableHead, tableBody
		pandoc2.table(block["c"])
	if block["t"] == "Div":
		# attr, block[]
		for newBlock in block["c"][1]:
			processBlock(newBlock, pandoc2)
	if block["t"] == "Null":
		pass



"""CLI Application
Cat a pandoc json string
"""

import json
import argparse
import platform
import ctypes
from cli2gui import Cli2Gui
import pypandoc
from catpandoc import pandoc2plain, pandoc2ansi, processpandoc


def handle(args):
	"""Handle the args and output to the terminal

	Args:
		args (argparse.Namespace): Args
	"""
	# Fix terminal for windows
	if platform.system() == "Windows":
		kernel32 = ctypes.windll.kernel32
		kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)

	# Open file and convert to JSON
	try:
		pypandoc.convert_text("#test", "json", format='md')
	except (FileNotFoundError, OSError):
		pypandoc.download_pandoc()
	output = json.loads(pypandoc.convert_file(args.file, 'json'))

	# Process args
	width = 80
	padding = 0
	if args.width is not None:
		width = int(args.width)
	if args.padding is not None:
		width = int(args.padding)
	theme = (4, 0, 4)
	if args.theme is not None:
		theme = tuple([int(col) for col in args.theme.split(",")])

	# Print to console
	if args.to_plain:
		for block in output["blocks"]:
			pandoc = pandoc2plain.Pandoc2Plain(width, padding)
			processpandoc.processBlock(block, pandoc)
			print(pandoc.genOutput())
	else:
		for block in output["blocks"]:
			pandoc = pandoc2ansi.Pandoc2Ansi(width, padding, theme)
			processpandoc.processBlock(block, pandoc)
			print(pandoc.genOutput())


@Cli2Gui(run_function=handle)
def cli():
	"""Parse args from the command line
	"""
	parser = argparse.ArgumentParser(description="Print md")
	parser.add_argument("file", help="file to render")
	parser.add_argument("--width", help="terminal width", action="store")
	parser.add_argument("--padding", help="terminal padding", action="store")
	parser.add_argument("--theme", help="theme to use r,g,b 0-5", action="store")
	parser.add_argument("--to-ansi", help="convert to ansi", action="store_true")
	parser.add_argument("--to-plain", help="convert to plaintext", action="store_true")
	args = parser.parse_args()
	handle(args)

if __name__ == "__main__":
	cli()



"""Do setup for uploading to pypi
"""
import setuptools

with open("README.md", "r") as readme:
	long_description = readme.read()

setuptools.setup(
	name="catpandoc",
	version="2020",
	author="FredHappyface",
	description="Cat multiple document files to the terminal",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/FHPythonUtils/CatPandoc",
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
		"Operating System :: OS Independent",
	],
	entry_points={
        'console_scripts': [
            'catpandoc=catpandoc.application:cli',
        ],
    },
	python_requires='>=3.0',
)



"""Convert a pandoc string to plaintext
"""
import urllib
import os
import catimage
import art
from catpandoc import processpandoc

class Pandoc2Plain():
	"""Convert a pandoc string to plaintext
	"""
	def __init__(self, width=80, padding=0):
		self.content = []
		self.width = width - padding * 2
		self.padding = padding


	#########################################
	# UTIL
	#########################################

	def newline(self):
		'''newline '''
		self.print()

	def catImage(self, content):
		'''catImage '''
		try:
			if content.startswith("http"):
				urllib.request.urlretrieve(content, "dowloadedImage")
				content = "dowloadedImage"
			try:
				self.print(catimage.generateGreyscale(os.getcwd() + os.sep + content,
				self.width))
			except FileNotFoundError:
				self.print(content)
		except urllib.error.HTTPError:
			self.print(content)


	def print(self, *args, end="\n"):
		'''Define a custom print method that has a very similar signature
		to the inbuilt print method '''
		self.content.append(" ".join([str(text) for text in args]) + end)

	def genOutput(self):
		'''Generate output '''
		lines = "".join(self.content).split("\n")
		out = []
		for line in lines:
			out.append(" " * self.padding + line)
		return "\n".join(out)

	#########################################
	# BLOCK
	#########################################

	def header(self, content):
		'''Process a header '''
		self.print("\n", end=" ")
		concatContent = ""
		if content[0] > 3:
			for part in content[-1]:
				processpandoc.processInline(part, self)
		else:
			concatContent = "".join([processpandoc.toPlaintext(part) for part in content[-1]])
			if any(char in concatContent for char in ["q", "y", "p", "g", "j"]):
				limit = 1
			else:
				limit = 2
			if content[0] == 1:
				if len(concatContent) > self.width / 7:
					concatContent = concatContent[:int(self.width / 7)]
				self.print("\n".join(art.text2art(concatContent, "swan").split("\n")[2:-limit-1]))
				self.print("▀"*self.width)
			if content[0] == 2:
				if len(concatContent) > self.width / 5:
					concatContent = concatContent[:int(self.width / 5)]
				self.print("\n".join(art.text2art(concatContent, "thin").split("\n")[1:-limit]))
				self.print("━"*self.width)
			if content[0] == 3:
				if len(concatContent) > self.width / 3:
					concatContent = concatContent[:int(self.width / 3)]
				self.print("\n".join(art.text2art(concatContent, "cygnet").split("\n")[1:-limit]))
				self.print("─"*self.width)

	def codeBlock(self, content):
		'''Process a code block  '''
		self.print("\n │ ", end="")
		self.print("\n │ ".join(content[1].split("\n")))

	def definitionList(self, content):
		'''Process a definition list '''
		for definition in content:
			#for definition in definitionBlock:
			self.newline()
			for part in definition[0]:
				processpandoc.processInline(part, self)
			self.print("\t:\t:\t", end="")
			for part in definition[1]:
				for partB in part:
					processpandoc.processBlock(partB, self)


	def orderedList(self, content):
		'''Process an ordered list '''
		for index, bullet in enumerate(content[1]):
			self.print("\n", index+1, end=". ")
			for point in bullet:
				if point["t"] in ["BulletList"]:
					self.print(" > ", end="")
				processpandoc.processBlock(point, self)

	def bulletList(self, content):
		'''Process a bulleted list '''
		for bullet in content:
			for point in bullet:
				if point["t"] not in ["BulletList"]:
					self.print("\n- ", end="")
				else:
					self.print(" > ", end="")
				processpandoc.processBlock(point, self)

	def table(self, content):
		'''Process a table '''
		# inline[], alignment[], double[], tablecell[], tablecell[][]
		# ignore. align, ignore, tableHead, tableBody
		colWidth = str(int(self.width/len(content[1]))-1)
		alignment = {"AlignLeft": "{:<"+colWidth+"."+colWidth+"}",
		"AlignCenter": "{:^"+colWidth+"."+colWidth+"}",
		"AlignRight": "{:>"+colWidth+"."+colWidth+"}",
		"AlignDefault": "{:<"+colWidth+"."+colWidth+"}"}
		self.print()
		self.print("│", end="")
		for index, tableHead in enumerate(content[3]):
			headContent = "".join([processpandoc.toPlaintext(headPart) for headPart in tableHead])
			self.print((alignment[content[1][index]["t"]]).format(headContent) + "│", end="")
		self.print("\n│", end="")
		for index, _ in enumerate(content[1]):
			self.print((alignment[content[1][index]["t"]]).format("─"*self.width) + "┤", end="")
		for tableRow in content[4]:
			self.print("\n│", end="")
			for index, tableCol in enumerate(tableRow):
				colContent = "".join([processpandoc.toPlaintext(colPart) for colPart in tableCol])
				self.print((alignment[content[1][index]["t"]]).format(colContent) + "│", end="")
		self.print()


	def blockQuote(self, content):
		'''Process a block quote '''
		print("\n ┃ ", end="")
		print("\n ┃ ".join([processpandoc.toPlaintext(block).split("\n") for block in content][0]))


	#########################################
	# INLINE
	#########################################

	def space(self):
		'''Process a space '''
		self.print(" ", end="")

	def emph(self, content):
		'''Process emphasized text '''
		for newInline in content:
			self.print("_", end="")
			processpandoc.processInline(newInline, self)
			self.print("_", end="")

	def strong(self, content):
		'''Process strong (bold) text '''
		for newInline in content:
			self.print("*", end="")
			processpandoc.processInline(newInline, self)
			self.print("*", end="")

	def strikeout(self, content):
		'''Process strikeout (crossed out) text '''
		for newInline in content:
			self.print("-", end="")
			processpandoc.processInline(newInline, self)
			self.print("-", end="")

	def superscript(self, content):
		'''Process superscript text '''
		for newInline in content:
			processpandoc.processInline(newInline, self)

	def subscript(self, content):
		'''Process subscript text '''
		for newInline in content:
			processpandoc.processInline(newInline, self)

	def smallCaps(self, content):
		'''Process small caps text '''
		for newInline in content:
			processpandoc.processInline(newInline, self)

	def quoted(self, content):
		'''Process quoted text '''
		for newInline in content[1]:
			self.print("\'", end="")
			processpandoc.processInline(newInline, self)
			self.print("\'", end="")

	def code(self, content):
		'''Process code '''
		self.print(" │" + content[1], end="│ ")

	def math(self, content):
		'''Process math '''
		self.print("Math: " + content[1])


	def note(self, content):
		'''Process a note '''
		for block in content:
			processpandoc.processBlock(block, self)


	def span(self, content):
		'''Process a span '''
		for newInline in content[1]:
			processpandoc.processInline(newInline, self)


	def image(self, content):
		'''Process an image '''
		self.catImage(content[-1][0])

	def link(self, content):
		'''Process a link '''
		for part in content[1]:
			processpandoc.processInline(part, self)
		self.print(" -> " + content[2][0], end="")

	def hr(self):
		'''Process a hr '''
		self.print("─"*self.width)
